#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

from collections import defaultdict
import csv
import io
import os
import re
import tarfile
import time
import treelib
from typing import Union

NAME_PREFIX = 'LINK-'
NAME_RE = r'^%s%s$' % (NAME_PREFIX, '[A-F0-9]{26,26}')
NAME_RE_COMPAT = r'^(\d{4,4}|LINK)-[0-9A-F]{26,26}$'


class UtilError(Exception):
    pass


def s1_in_s2(s1: str, s2: Union[str, list[str]]) -> bool:
    if isinstance(s2, str):
        return s1 == s2
    elif isinstance(s2, list):
        return s1 in s2
    else:
        raise ValueError('Invalid type for s2. '
                         'Must be a string or a list of strings.')


def is_writable(path: str) -> bool:
    # Check if the file exists
    if os.path.exists(path):
        # If the file exists, check if it is writable
        return os.access(path, os.W_OK)
    else:
        # If the file doesn't exist, check if the directory is writable
        parent_dir = os.path.dirname(path) or '.'
        return os.access(parent_dir, os.W_OK)


def read_cert_archive(*, path: str) -> dict[str, tuple[str, str]]:
    def parse_name(name):
        pat = (r'^(intermediate|root)/'
               r'[0-9A-F]{64,64}\.pem$')
        if not re.search(pat, name):
            e = 'malformed path in archive: %s' % name
            raise UtilError(e)
        type_, pem = os.path.split(name)
        sha256 = pem[:64]

        return type_, sha256

    data = {}
    try:
        with tarfile.open(name=path, mode='r') as tar:
            for member in tar:
                if member.name in ['root', 'intermediate']:
                    continue
                cert_type, sha256 = parse_name(member.name)
                f = tar.extractfile(member)
                content = f.read()
                data[sha256] = (cert_type, content)

    except (tarfile.TarError, OSError) as e:
        raise UtilError(str(e))

    return data


def write_cert_archive(*, path: str, data: dict[str, tuple[str, str]]):
    try:
        with tarfile.open(name=path, mode='w:gz') as tar:
            for k, v in data.items():
                name = os.path.join(v[0], k + '.pem')
                member = tarfile.TarInfo(name=name)
                member.size = len(v[1])
                member.mtime = time.time()
                content = (v[1].encode() if isinstance(v[1], str)
                           else v[1])
                f = io.BytesIO(content)
                tar.addfile(member, fileobj=f)
    except (tarfile.TarError, OSError) as e:
        raise UtilError(str(e))


def hash_to_name(*, sha256: str) -> str:
    # PAN-OS certificate-name max len 63
    # Panorama certificate-name max len 31
    # PAN-99186 won't do
    x = NAME_PREFIX + sha256
    return x[0:31]


def read_fingerprints(*, path: str) -> list[dict[str, str]]:
    try:
        with open(path, 'r', newline='') as csvfile:
            reader = csv.DictReader(csvfile,
                                    dialect='unix')
            x = []
            for row in reader:
                x.append(row)
    except OSError as e:
        raise UtilError(str(e))

    return x


def write_fingerprints(*, path: str, data: list[dict[str, str]]):
    fieldnames = [
        'type',
        'sha256',
    ]

    try:
        with open(path, 'w', newline='') as csvfile:
            writer = csv.DictWriter(csvfile,
                                    dialect='unix',
                                    fieldnames=fieldnames)
            writer.writeheader()
            for x in data:
                row = {
                    'type': x['type'],
                    'sha256': x['sha256'],
                }
                writer.writerow(row)

    except OSError as e:
        raise UtilError(str(e))


def tree_to_dict(*, tree: treelib.Tree) -> dict:
    nodes = []

    for node in tree.all_nodes():
        parent = tree.parent(node.identifier)
        parent = parent if parent is None else parent.identifier
        x = {
            'identifier': node.identifier,
            'tag': node.tag,
            'data': node.data,
            'parent': parent,
        }
        nodes.append(x)

    return {'nodes': nodes}


def dict_to_tree(*, data: dict) -> treelib.Tree:
    root = {
        'identifier': 0,
        'tag': 'Root',
        'parent': None,
        'data': None,
    }

    if ('nodes' not in data or
       data['nodes'][0] != root):
        raise UtilError('Malformed tree dict')

    tree = treelib.Tree()

    for x in data['nodes']:
        tree.create_node(
            identifier=x['identifier'],
            tag=x['tag'],
            parent=x['parent'],
            data=x['data'])

    return tree


def stats_from_tree(*, tree: treelib.Tree) -> dict[str, Union[float, int]]:
    all_nodes = tree.all_nodes()
    root_nodes = [node for node in all_nodes
                  if tree.parent(node.identifier) is not None and
                  tree.parent(node.identifier).identifier == 0]
    roots_with_no_children = [node for node in root_nodes
                              if not len(tree.children(node.identifier))]
    children_counts = [len(tree.children(node.identifier))
                       for node in all_nodes]
    node_depths = [tree.depth(node.identifier) for node in all_nodes]
    level_counts = defaultdict(int)
    for x in node_depths:
        level_counts[x] += 1

    stats = {
        'total_nodes': len(all_nodes),  # includes root node
        'total_roots': len(root_nodes),
        'roots_with_no_children': len(roots_with_no_children),
        'total_intermediates': len(all_nodes) - 1 - len(root_nodes),
        'maximum_depth': max(node_depths, default=0),
        'average_depth':
        sum(node_depths) / len(node_depths) if node_depths else 0,
        'maximum_breadth': max(level_counts.values(), default=0),
        'maximum_children': max(children_counts, default=0),
        'average_children':
        sum(children_counts) / len(children_counts) if children_counts else 0,
        'nodes_with_10+_children': sum(1 for c in children_counts if c >= 10),
        'nodes_with_50+_children': sum(1 for c in children_counts if c >= 50),
        'nodes_with_100+_children':
        sum(1 for c in children_counts if c >= 100),
        'leaf_nodes': sum(1 for c in children_counts if c == 0),
    }

    return stats



#!/usr/bin/env python3

#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import argparse
import asyncio
from collections import defaultdict
from datetime import datetime, timezone
from html import escape
import json
import os
import pprint
import sys

libpath = os.path.dirname(os.path.abspath(__file__))
sys.path[:0] = [os.path.join(libpath, os.pardir)]

from pan_chainguard import title, __version__
from pan_chainguard.ccadb import root_status_bits_flag, root_status_bits
import pan_chainguard.util

args = None


def root_status(node):
    if not args.verbose or node.tag == 'Root':
        return ''

    data = node.data
    if data['Certificate Record Type'] == 'Root Certificate':
        bits = root_status_bits_flag(data)
        status = root_status_bits(bits, compact=True)
        return status
    else:
        return ''


def tree_sort(node):
    return str(node.data['Certificate Name'])


def format_text(tree):
    txt = tree.show(stdout=None, key=tree_sort)
    print(txt, end='')


def format_stats(tree):
    stats = pan_chainguard.util.stats_from_tree(tree=tree)

    for k, v in stats.items():
        name = k.replace('_', ' ').title()
        value = '%.4f' % v if isinstance(v, float) else '%d' % v
        print('%s: %s' % (name, value))


def format_rst(tree):
    def tree_to_rst(tree, node_id=None, level=-1):
        if node_id is None:
            node_id = tree.root

        node = tree[node_id]
        lines = []

        sha256 = str(node_id)
        # skips root node
        if len(sha256) == 64:
            # XXX uncertain if we can monospace the anchor
            lines.append(
                f'{"  " * level}* '
                f'`{sha256} <https://crt.sh/?sha256={sha256}>`_ '
                f'{node.tag[64:]}'
            )

        for i, child in enumerate(sorted(tree.children(node_id),
                                         key=tree_sort)):
            if i == 0:
                lines.append('')
            lines.extend(tree_to_rst(tree, child.identifier, level + 1))

        return lines

    lines = tree_to_rst(tree)
    rst = ''
    if args.title:
        rst = f'{args.title}\n{"=" * len(args.title)}\n'
    rst += '\n'.join(lines)
    print(rst)


def format_html(tree):
    def tree_to_html(tree, node_id=None):
        if node_id is None:
            node_id = tree.root

        node = tree[node_id]
        children = tree.children(node_id)
        root_vendors = root_status(node)
        if root_vendors:
            root_vendors = f' <b>vendors:{root_vendors}</b> '
        html = ''

        sha256 = str(node_id)
        # skips root node
        if len(sha256) == 64:
            if node.data['Certificate Record Type'] == 'Root Certificate':
                try:
                    tree_to_html.roots += 1
                except AttributeError:
                    tree_to_html.roots = 1
            else:
                try:
                    tree_to_html.intermediates += 1
                except AttributeError:
                    tree_to_html.intermediates = 1

            html += (f'<li><a href="https://crt.sh/?sha256={sha256}">'
                     f'<code>{sha256}</code></a>'
                     f'{root_vendors}'
                     f'{escape(node.tag[64:])}</li>\n')

        if children:
            html += '<ul>\n'
            for child in sorted(children, key=tree_sort):
                html += tree_to_html(tree, child.identifier)
            html += '</ul>\n'

        return html

    tree_html = tree_to_html(tree)

    html = ''
    if args.title:
        html += f'<h1>{escape(args.title)}</h1>\n'

    if args.verbose:
        html += f'''<h2>Certificate Totals</h2>
<blockquote>
Root Certificates: {tree_to_html.roots}<br>
Intermediate Certificates: {tree_to_html.intermediates}<br>
</blockquote>
'''
    if args.verbose:
        html += '<h2>Certificate Tree</h2>\n'
    html += tree_html
    if args.verbose:
        stats = pan_chainguard.util.stats_from_tree(tree=tree)
        stats_ = ''
        for k, v in stats.items():
            name = k.replace('_', ' ').title()
            value = '%.4f' % v if isinstance(v, float) else '%d' % v
            stats_ += '%s: %s<br>\n' % (name, value)
        html += f'''<h2>Certificate Tree Statistics</h2>
<blockquote>
{stats_}</blockquote>
'''

        now_utc = datetime.now(timezone.utc)
        date_str = now_utc.strftime('%Y-%m-%d %H:%M:%S UTC')
        html += f'''<footer>
<p><em>Generated: {date_str}</em></p>
</footer>
'''

    print(html, end='')


def format_json(tree):
    data = pan_chainguard.util.tree_to_dict(tree=tree)
    json_data = json.dumps(data, indent=4)
    print(json_data)


formats = {
    'txt': format_text,
    'rst': format_rst,
    'html': format_html,
    'json': format_json,
    'stats': format_stats,
}


def main():
    global args
    args = parse_args()

    ret = asyncio.run(main_loop())

    sys.exit(ret)


async def main_loop():
    tree = read_tree()

    if args.test_collisions:
        if not test_collisions(tree):
            return 1

    if args.format:
        for format in args.format:
            formats[format](tree)

    if args.fingerprint:
        for x in args.fingerprint:
            lookup(tree, x)

    return 0


def read_tree():
    try:
        with open(args.tree, 'r') as f:
            json_data = f.read()
        data = json.loads(json_data)
    except (OSError, ValueError) as e:
        print('%s: %s' % (args.tree, e), file=sys.stderr)
        sys.exit(1)

    try:
        tree = pan_chainguard.util.dict_to_tree(data=data)
    except pan_chainguard.util.UtilError as e:
        print('%s: %s' % (args.tree, e), file=sys.stderr)
        sys.exit(1)

    return tree


# The first 26 characters of the SHA-256 fingerprint (length 64) are
# used for the PAN-OS certificate name; test name for collisions.
def test_collisions(tree):
    data = pan_chainguard.util.tree_to_dict(tree=tree)

    names = defaultdict(list)
    for node in data['nodes']:
        ident = node.get('identifier')
        if ident:
            name = pan_chainguard.util.hash_to_name(sha256=ident)
            names[name].append(ident)

    # keep only names that map to >1 identifiers
    collisions = {name: ids for name, ids in names.items() if len(ids) > 1}

    if collisions:
        print(f'{len(collisions)} certificate name collisions',
              file=sys.stderr)
        for name in sorted(collisions):
            for ident in collisions[name]:
                print(f'    {name} {ident}', file=sys.stderr)
        return False
    else:
        print('no certificate name collisions', file=sys.stderr)
        return True


def lookup(tree, sha256):
    node = tree.get_node(sha256)

    if node is None:
        print('Not found: %s' % sha256, file=sys.stderr)
        return

    data = node.data
    filtered_data = {k: v for k, v in data.items() if v != ''}

    print(pprint.pformat(filtered_data))


def parse_args():
    parser = argparse.ArgumentParser(
        usage='%(prog)s [options]',
        description='certificate tree analysis and reporting')
    parser.add_argument('--tree',
                        required=True,
                        metavar='PATH',
                        help='JSON certificate tree path')
    parser.add_argument('-f', '--format',
                        action='append',
                        choices=formats.keys(),
                        help='output format')
    parser.add_argument('-t', '--title',
                        help='report title')
    parser.add_argument('--test-collisions',
                        action='store_true',
                        help='test for certificate name collisions')
    parser.add_argument('-F', '--fingerprint',
                        action='append',
                        metavar='SHA-256',
                        help='lookup CCADB data by certificate SHA-256 '
                        'fingerprint')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='enable verbosity')
    parser.add_argument('--debug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='enable debug')
    x = '%s %s' % (title, __version__)
    parser.add_argument('--version',
                        action='version',
                        help='display version',
                        version=x)
    args = parser.parse_args()

    if args.debug:
        print(args, file=sys.stderr)

    return args


if __name__ == '__main__':
    main()



#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

# https://www.ccadb.org/resources
#
# All Certificate Information (root and intermediate) in CCADB (CSV)
# utility functions.
# https://ccadb.my.salesforce-sites.com/ccadb/AllCertificateRecordsCSVFormatv2
#
# All Included Root Certificate Trust Bit Settings:
# https://ccadb.my.salesforce-sites.com/ccadb/AllIncludedRootCertsCSV

import csv
from datetime import datetime, timezone
from enum import Flag, auto
import sys
from typing import Tuple, Union, Optional


class CcadbError(Exception):
    pass


_FMT = '%Y.%m.%d %z'

__all__ = [
    'valid_from', 'valid_to', 'valid_from_to', 'revoked',
    'TrustBits', 'trust_bits_flag',
    'derived_trust_bits_list', 'derived_trust_bits',
    'root_trust_bits_list', 'root_trust_bits',
    'RootStatusBits', 'root_status_bits_flag', 'root_status_bits',
]


class TrustBits(Flag):
    NONE = 0
    OTHER = auto()
    CLIENT_AUTHENTICATION = auto()
    CODE_SIGNING = auto()
    DOCUMENT_SIGNING = auto()
    OCSP_SIGNING = auto()
    SECURE_EMAIL = auto()
    SERVER_AUTHENTICATION = auto()
    TIME_STAMPING = auto()


TrustBitsMap = {
    'Client Authentication': TrustBits.CLIENT_AUTHENTICATION,
    'Code Signing': TrustBits.CODE_SIGNING,
    'Document Signing': TrustBits.DOCUMENT_SIGNING,
    'OCSP Signing': TrustBits.OCSP_SIGNING,
    'Secure Email': TrustBits.SECURE_EMAIL,
    'Server Authentication': TrustBits.SERVER_AUTHENTICATION,
    'Time Stamping': TrustBits.TIME_STAMPING,
}

TrustBitsMap2 = {v.name: v for v in TrustBitsMap.values()}


class RootStatusBits(Flag):
    NONE = 0
    MOZILLA = auto()
    CHROME = auto()
    APPLE = auto()
    MICROSOFT = auto()


RootStatusBitsMap = {
    'Mozilla Status': RootStatusBits.MOZILLA,
    'Apple Status': RootStatusBits.APPLE,
    'Chrome Status': RootStatusBits.CHROME,
    'Microsoft Status': RootStatusBits.MICROSOFT,
}


def _now():
    now = datetime.now(timezone.utc)

    return now


# CCADB Valid From/To fields are the date only in YYYY.MM.DD format.
# When converted to a date with time, the time used is midnight; for
# example:
#   2024.10.27 -> 2024-10-27 00:00:00+00:00

def valid_from(row: dict[str, str]) -> Tuple[bool, Union[str, None]]:
    k = 'Valid From (GMT)'
    valid_from = datetime.strptime(row[k] + ' +0000', _FMT)
    if _now() < valid_from:
        x = 'Not yet valid (valid from %s)' % row[k]
        return False, x

    return True, None


def valid_to(row: dict[str, str]) -> Tuple[bool, Union[str, None]]:
    k = 'Valid To (GMT)'
    valid_to = datetime.strptime(row[k] + ' +0000', _FMT)
    if valid_to < _now():
        x = 'Expired (valid to %s)' % row[k]
        return False, x

    return True, None


def valid_from_to(row: dict[str, str]) -> Tuple[bool, Union[str, None]]:
    return valid_from(row) and valid_to(row)


def revoked(row: dict[str, str]) -> Tuple[bool, Union[str, None]]:
    k = 'Revocation Status'
    if row[k] not in ['', 'Not Revoked']:
        return True, row[k]

    return False, None


def _trust_bits_list(key: str, row: dict[str, str]) -> list[str]:
    x = row[key]
    if not x:
        return []

    return x.split(';')


def derived_trust_bits_list(row: dict[str, str]) -> list[str]:
    type_ = 'Intermediate Certificate'
    x = row['Certificate Record Type']
    if x != type_:
        raise ValueError('certificate type "%s" not "%s"' % (x, type_))

    key = 'Derived Trust Bits'
    return _trust_bits_list(key, row)


def root_trust_bits_list(row: dict[str, str]) -> list[str]:
    type_ = 'Root Certificate'
    x = row['Certificate Record Type']
    if x != type_:
        raise ValueError('certificate type "%s" not "%s"' % (x, type_))

    key = 'Trust Bits for Root Cert'
    return _trust_bits_list(key, row)


def trust_bits_flag(values: list[str]) -> TrustBits:
    bits = TrustBits.NONE
    for x in values:
        if x in TrustBitsMap:
            bits = bits | TrustBitsMap[x]
        else:
            bits = bits | TrustBits.OTHER

    return bits


def derived_trust_bits(row: dict[str, str]) -> TrustBits:
    x = derived_trust_bits_list(row)
    return trust_bits_flag(x)


def root_trust_bits(row: dict[str, str]) -> TrustBits:
    x = root_trust_bits_list(row)
    return trust_bits_flag(x)


def root_status_bits_flag(row: dict[str, str]) -> RootStatusBits:
    type_ = 'Root Certificate'
    x = row['Certificate Record Type']
    if x != type_:
        raise ValueError('certificate type "%s" not "%s"' % (x, type_))

    bits = RootStatusBits.NONE
    for x in RootStatusBitsMap:
        if x in row and row[x] == 'Included':
            bits = bits | RootStatusBitsMap[x]

    return bits


map_ = {
    ('mozilla', 'Mz'): RootStatusBits.MOZILLA,
    ('apple', 'Ap'): RootStatusBits.APPLE,
    ('chrome', 'Ch'): RootStatusBits.CHROME,
    ('microsoft', 'Ms'): RootStatusBits.MICROSOFT,
}


def root_status_bits(bits: RootStatusBits,
                     compact: bool = False) -> Union[str, list]:
    r = []

    for k, v in map_.items():
        if v in bits:
            r.append(k[1] if compact else k[0])

    return ''.join(r) if compact else r


class CcadbRootTrustSettings:
    def __init__(self, *,
                 path: str,
                 debug: bool = False):
        self._certs = {}
        self._debug = debug

        try:
            with open(path, 'r', newline='') as csvfile:
                reader = csv.DictReader(csvfile,
                                        dialect='unix')
                for row in reader:
                    sha256 = row['SHA-256 Fingerprint']
                    if self._debug and sha256 in self.certs:
                        print('Duplicate in CCADB Root Trust Settings %s' % (
                            sha256), file=sys.stderr)
                    self._certs[sha256] = row

        except OSError as e:
            raise CcadbError(str(e))

    @property
    def certs(self):
        return self._certs

    def get(self, *, sha256: str) -> Optional[dict[str, str]]:
        if sha256 not in self._certs:
            return

        return self._certs[sha256]

    def root_status_bits_flag(self, *, sha256: str) -> Optional[
            RootStatusBits]:
        if sha256 not in self._certs:
            return
        row = self._certs[sha256]

        bits_map = RootStatusBitsMap.copy()
        del bits_map['Chrome Status']
        bits_map['Google Chrome Status'] = RootStatusBits.CHROME

        bits = RootStatusBits.NONE
        for x in bits_map:
            if x in row and row[x] == 'Included':
                bits = bits | bits_map[x]

        return bits

    # There is no 'Google Chrome Trust Bits'.
    # The Chrome Root Store only trusts Server Authentication.
    def chrome_trust_bits_list(self, *, sha256: str) -> Optional[list]:
        if sha256 not in self._certs:
            return

        bits = self.root_status_bits_flag(sha256=sha256)
        if RootStatusBits.CHROME not in bits:
            return []

        return ['Server Authentication']

    def chrome_trust_bits(self, *, sha256: str) -> Optional[TrustBits]:
        values = self.chrome_trust_bits_list(sha256=sha256)
        if values is None:
            return

        if not values:
            return TrustBits.NONE

        return TrustBits.SERVER_AUTHENTICATION

    def mozilla_trust_bits_list(self, *, sha256: str) -> Optional[list]:
        if sha256 not in self._certs:
            return

        return _trust_bits_list('Mozilla Trust Bits', self._certs[sha256])

    def mozilla_trust_bits(self, *, sha256: str) -> Optional[TrustBits]:
        values = self.mozilla_trust_bits_list(sha256=sha256)
        if values is None:
            return

        if self._debug and not values:
            print('Mozilla Trust Bits null %s' % sha256, file=sys.stderr)

        if values == ['All Trust Bits Turned Off']:
            return TrustBits.NONE

        MAP = {
            'Websites': TrustBits.SERVER_AUTHENTICATION,
            'Email': TrustBits.SECURE_EMAIL,
        }

        bits = TrustBits.NONE
        for x in values:
            if x in MAP:
                bits = bits | MAP[x]
            else:
                bits = bits | TrustBits.OTHER

        if self._debug and TrustBits.OTHER in bits:
            print('Mozilla Trust Bits other %s %s' % (
                values, sha256), file=sys.stderr)

        return bits

    def apple_trust_bits_list(self, *, sha256: str) -> Optional[list]:
        if sha256 not in self._certs:
            return

        return _trust_bits_list('Apple Trust Bits', self._certs[sha256])

    def apple_trust_bits(self, *, sha256: str) -> Optional[TrustBits]:
        values = self.apple_trust_bits_list(sha256=sha256)
        if values is None:
            return

        if self._debug and not values:
            print('Apple Trust Bits null %s' % sha256, file=sys.stderr)

        MAP = {
            'serverAuth': TrustBits.SERVER_AUTHENTICATION,
            'clientAuth': TrustBits.CLIENT_AUTHENTICATION,
            'emailProtection': TrustBits.SECURE_EMAIL,
            'timeStamping': TrustBits.TIME_STAMPING,
            'codeSigning': TrustBits.CODE_SIGNING,
            #  'BrandIndicatorforMessageIdentification': TrustBits.OTHER,
        }

        bits = TrustBits.NONE
        for x in values:
            if x in MAP:
                bits = bits | MAP[x]
            else:
                bits = bits | TrustBits.OTHER

        if self._debug and TrustBits.OTHER in bits:
            print('Apple Trust Bits other %s %s' % (
                values, sha256), file=sys.stderr)

        return bits

    def microsoft_trust_bits_list(self, *, sha256: str) -> Optional[list]:
        if sha256 not in self._certs:
            return

        # DTBs - Derived Trust Bits
        return _trust_bits_list('Microsoft EKUs', self._certs[sha256])

    def microsoft_trust_bits(self, *, sha256: str) -> Optional[TrustBits]:
        values = self.microsoft_trust_bits_list(sha256=sha256)
        if values is None:
            return

        if self._debug and not values:
            print('Microsoft Trust Bits null %s' % sha256, file=sys.stderr)

        bits = trust_bits_flag(values)

        return bits



import unittest

from pan_chainguard.ccadb import *
from pan_chainguard.ccadb import TrustBitsMap

REVOCATION_STATUS = 'Revocation Status'
DERIVED_TRUST_BITS = 'Derived Trust Bits'  # Intermediate Certificate
ROOT_TRUST_BITS = 'Trust Bits for Root Cert'  # Root Certificate
TRUST_BITS = {
    DERIVED_TRUST_BITS: {
        '': derived_trust_bits,
        'list': derived_trust_bits_list,
        'flag': trust_bits_flag,
        'type': 'Intermediate Certificate',
    },
    ROOT_TRUST_BITS: {
        '': root_trust_bits,
        'list': root_trust_bits_list,
        'flag': trust_bits_flag,
        'type': 'Root Certificate',
    },
}


class CcadbTest(unittest.TestCase):
    def test_01(self):
        x = ''
        t = {REVOCATION_STATUS: x}
        r, err = revoked(t)
        self.assertFalse(r, "%s: %s" % (x, err))
        self.assertIsNone(err)

    def test_02(self):
        x = 'Not Revoked'
        t = {REVOCATION_STATUS: x}
        r, err = revoked(t)
        self.assertFalse(r, "%s: %s" % (x, err))
        self.assertIsNone(err)

    def test_03(self):
        x = 'Revoked'
        t = {REVOCATION_STATUS: x}
        r, err = revoked(t)
        self.assertTrue(r, "%s: %s" % (x, err))
        self.assertIsNotNone(err)
        self.assertEqual('Revoked', err)

    def test_04(self):
        x = 'Unknown'
        t = {REVOCATION_STATUS: x}
        r, err = revoked(t)
        self.assertTrue(r, "%s: %s" % (x, err))
        self.assertIsNotNone(err)
        self.assertEqual('Unknown', err)

    def test_05(self):
        for k in TRUST_BITS:
            x = ''
            t = {
                'Certificate Record Type': TRUST_BITS[k]['type'],
                k: x,
            }

            r = TRUST_BITS[k]['list'](t)
            self.assertListEqual([], r)

            bits = TRUST_BITS[k]['flag'](r)
            self.assertEqual(TrustBits.NONE, bits)

            bits = TRUST_BITS[k][''](t)
            self.assertEqual(TrustBits.NONE, bits)

    def test_06(self):
        for k in TRUST_BITS:
            x = 'abcd'
            t = {
                'Certificate Record Type': TRUST_BITS[k]['type'],
                k: x,
            }

            r = TRUST_BITS[k]['list'](t)
            self.assertListEqual([x], r)

            bits = TRUST_BITS[k]['flag'](r)
            self.assertEqual(TrustBits.OTHER, bits)

            bits = TRUST_BITS[k][''](t)
            self.assertEqual(TrustBits.OTHER, bits)

    def test_07(self):
        for k in TRUST_BITS:
            x = 'Server Authentication'
            t = {
                'Certificate Record Type': TRUST_BITS[k]['type'],
                k: x,
            }

            r = TRUST_BITS[k]['list'](t)
            self.assertListEqual([x], r)

            bits = TRUST_BITS[k]['flag'](r)
            self.assertEqual(TrustBitsMap[x], bits)

            bits = TRUST_BITS[k][''](t)
            self.assertEqual(TrustBitsMap[x], bits)

    def test_08(self):
        for k in TRUST_BITS:
            x = ';'.join(TrustBitsMap.keys())
            t = {
                'Certificate Record Type': TRUST_BITS[k]['type'],
                k: x,
            }

            r = TRUST_BITS[k]['list'](t)
            self.assertListEqual(list(TrustBitsMap.keys()), r)

            bits = TRUST_BITS[k]['flag'](r)
            for x in TrustBitsMap.values():
                self.assertIn(x, bits)

            bits = TRUST_BITS[k][''](t)
            for x in TrustBitsMap.values():
                self.assertIn(x, bits)

    def test_09(self):
        row = {
            "Certificate Record Type": "Intermediate Certificate",
            "Mozilla Status": "Not Included",
            "Apple Status": "Not Included",
            "Chrome Status": "Not Included",
            "Microsoft Status": "Not Included",
        }
        r = RootStatusBits.NONE

        with self.assertRaises(ValueError) as e:
            bits = root_status_bits_flag(row)
        x = ('certificate type "Intermediate Certificate" '
             'not "Root Certificate"')
        self.assertEqual(str(e.exception), x)

    def test_10(self):
        row = {
            "Certificate Record Type": "Root Certificate",
            "Mozilla Status": "Not Included",
            "Apple Status": "Not Included",
            "Chrome Status": "Not Included",
            "Microsoft Status": "Not Included",
        }
        r = RootStatusBits.NONE

        bits = root_status_bits_flag(row)
        self.assertEqual(bits, r)
        r = root_status_bits(bits)
        self.assertEqual(r, [])

    def test_11(self):
        row = {
            "Certificate Record Type": "Root Certificate",
            "Mozilla Status": "Included",
            "Apple Status": "Included",
            "Chrome Status": "Included",
            "Microsoft Status": "Included",
        }
        r = (RootStatusBits.MOZILLA |
             RootStatusBits.APPLE |
             RootStatusBits.CHROME |
             RootStatusBits.MICROSOFT)

        bits = root_status_bits_flag(row)
        self.assertEqual(bits, r)
        r = root_status_bits(bits)
        self.assertEqual(r, ['mozilla', 'apple', 'chrome', 'microsoft'])
        r = root_status_bits(bits, compact=True)
        self.assertEqual(r, 'MzApChMs')

    def test_12(self):
        row = {
            "Certificate Record Type": "Root Certificate",
            "Mozilla Status": "Included",
            "Apple Status": "Included",
            "Chrome Status": "Not Included",
            "Microsoft Status": "Not Included",
        }
        r = (RootStatusBits.MOZILLA |
             RootStatusBits.APPLE)

        bits = root_status_bits_flag(row)
        self.assertEqual(bits, r)
        r = root_status_bits(bits)
        self.assertEqual(r, ['mozilla', 'apple'])
        r = root_status_bits(bits, compact=True)
        self.assertEqual(r, 'MzAp')



from datetime import timedelta
import unittest

from pan_chainguard.ccadb import *
from pan_chainguard.ccadb import _now

FMT = '%Y.%m.%d'
VALID_FROM = 'Valid From (GMT)'
VALID_TO = 'Valid To (GMT)'

now = _now()
today = now.strftime(FMT)
x = now - timedelta(days=1)
yesterday = x.strftime(FMT)
x = now + timedelta(days=1)
tomorrow = x.strftime(FMT)


class CcadbTest(unittest.TestCase):
    def test_01(self):
        x = today
        t = {VALID_FROM: x}
        r, err = valid_from(t)
        self.assertTrue(r, "%s: %s" % (x, err))
        self.assertIsNone(err)

    def test_02(self):
        x = yesterday
        t = {VALID_FROM: x}
        r, err = valid_from(t)
        self.assertTrue(r, "%s: %s" % (x, err))
        self.assertIsNone(err)

    def test_03(self):
        x = tomorrow
        t = {VALID_FROM: x}
        r, err = valid_from(t)
        self.assertFalse(r, "%s: %s" % (x, err))
        self.assertIsNotNone(err)
        self.assertIn('Not yet valid', err)

    def test_04(self):
        x = tomorrow
        t = {VALID_TO: x}
        r, err = valid_to(t)
        self.assertTrue(r, "%s: %s" % (x, err))
        self.assertIsNone(err)

    def test_05(self):
        x = yesterday
        t = {VALID_TO: x}
        r, err = valid_to(t)
        self.assertFalse(r, "%s: %s" % (x, err))
        self.assertIsNotNone(err)
        self.assertIn('Expired', err)

    def test_06(self):
        x = today
        t = {VALID_TO: x}
        r, err = valid_to(t)
        self.assertFalse(r, "%s: %s" % (x, err))
        self.assertIsNotNone(err)
        self.assertIn('Expired', err)

    def test_07(self):
        t = {
            VALID_FROM: yesterday,
            VALID_TO: tomorrow,
        }
        r, err = valid_from_to(t)
        self.assertTrue(r, "%s: %s" % (t, err))
        self.assertIsNone(err)

    def test_08(self):
        t = {
            VALID_FROM: yesterday,
            VALID_TO: today,
        }
        r, err = valid_from_to(t)
        self.assertFalse(r, "%s: %s" % (t, err))
        self.assertIsNotNone(err)
        self.assertIn('Expired', err)



#!/usr/bin/env python3

#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import argparse
import asyncio
from collections import defaultdict
import logging
import os
import pprint
import re
import sys
import time
from treelib import Node, Tree
import xml.etree.ElementTree as etree

try:
    import pan.xapi
except ImportError:
    print('Install pan-python: https://pypi.org/project/pan-python/',
          file=sys.stderr)
    sys.exit(1)

libpath = os.path.dirname(os.path.abspath(__file__))
sys.path[:0] = [os.path.join(libpath, os.pardir)]

from pan_chainguard import title, __version__
import pan_chainguard.util

# XXX keep compatible with sequence based naming for now
NAME_RE = pan_chainguard.util.NAME_RE_COMPAT

args = None


class Xpath():
    def __init__(self, *,
                 panorama=False,
                 template=None,
                 vsys=None):
        self.type = 'panorama' if panorama else 'firewall'
        if self.type == 'firewall' and template is not None:
            raise ValueError('template set for firewall')
        self.template = template
        self.vsys = vsys

    def __str__(self):
        x = self.__dict__
        return ', '.join((': '.join((k, str(x[k]))))
                         for k in sorted(x))

    @property
    def panorama(self):
        return self.type == 'panorama'

    def _variant(self, xpath):
        if self.type == 'firewall':
            if self.vsys is None:
                x = xpath[self.type]['shared']
            else:
                x = xpath[self.type]['vsys'] % self.vsys
        elif self.type == 'panorama':
            if self.template is None:
                x = xpath[self.type]['panorama']
            elif self.vsys is None:
                x = xpath[self.type]['template']['shared'] % self.template
            else:
                x = xpath[self.type]['template']['vsys'] % (
                    self.template, self.vsys)
        else:
            assert False, 'Malformed type: %s' % self.type

        return x

    def trusted_root_ca(self):
        xpath = {
            'firewall': {
                'shared': '/config/shared/ssl-decrypt/trusted-root-CA',
                'vsys': ("/config/devices/entry"
                         "[@name='localhost.localdomain']"
                         "/vsys/entry[@name='%s']/ssl-decrypt"
                         "/trusted-root-CA")
            },

            'panorama': {
                'panorama': '/config/panorama/ssl-decrypt/trusted-root-CA',
                'template': {
                    'shared': ("/config/devices/entry"
                               "[@name='localhost.localdomain']"
                               "/template/entry[@name='%s']/config/shared"
                               "/ssl-decrypt/trusted-root-CA"),
                    'vsys': ("/config/devices/entry"
                             "[@name='localhost.localdomain']"
                             "/template/entry[@name='%s']/config/devices"
                             "/entry[@name='localhost.localdomain']/vsys"
                             "/entry[@name='%s']/ssl-decrypt/trusted-root-CA")
                },
            },
        }

        return self._variant(xpath)

    def certificates(self):
        xpath = {
            'firewall': {
                'shared': '/config/shared/certificate',
                'vsys': ("/config/devices/entry[@name='localhost.localdomain']"
                         "/vsys/entry[@name='%s']/certificate"),
            },

            'panorama': {
                'panorama': '/config/panorama/certificate',
                'template': {
                    'shared': ("/config/devices/entry"
                               "[@name='localhost.localdomain']"
                               "/template/entry[@name='%s']"
                               "/config/shared/certificate"),
                    'vsys': ("/config/devices/entry"
                             "[@name='localhost.localdomain']"
                             "/template/entry[@name='%s']/config/devices"
                             "/entry[@name='localhost.localdomain']"
                             "/vsys/entry[@name='%s']/certificate")
                },
            },
        }

        return self._variant(xpath)

    def root_ca_exclude_list(self):
        xpath = {
            'firewall': {
                'shared': '/config/shared/ssl-decrypt/root-ca-exclude-list',
                'vsys': ("/config/devices/entry[@name='localhost.localdomain']"
                         "/vsys/entry[@name='%s']/ssl-decrypt/"
                         "root-ca-exclude-list"),
            },

            'panorama': {
                'panorama': ('/config/panorama/ssl-decrypt/'
                             'root-ca-exclude-list'),
                'template': {
                    'shared': ("/config/devices/entry"
                               "[@name='localhost.localdomain']"
                               "/template/entry[@name='%s']"
                               "/config/shared/ssl-decrypt/"
                               "root-ca-exclude-list"),
                    'vsys': ("/config/devices/entry"
                             "[@name='localhost.localdomain']"
                             "/template/entry[@name='%s']/config/devices"
                             "/entry[@name='localhost.localdomain']"
                             "/vsys/entry[@name='%s']/ssl-decrypt/"
                             "root-ca-exclude-list")
                },
            },
        }

        return self._variant(xpath)


def main():
    global args
    args = parse_args()

    logger = logging.getLogger(pan.xapi.__name__)
    if args.xdebug == 3:
        logger.setLevel(pan.xapi.DEBUG3)
    elif args.xdebug == 2:
        logger.setLevel(pan.xapi.DEBUG2)
    elif args.xdebug == 1:
        logger.setLevel(pan.xapi.DEBUG1)

    log_format = '%(name)s: %(message)s'
    handler = logging.StreamHandler()
    formatter = logging.Formatter(log_format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    asyncio.run(main_loop())

    sys.exit(0)


async def main_loop():
    xapi = None
    panorama = False

    try:
        xapi = pan.xapi.PanXapi(tag=args.tag,
                                debug=args.xdebug)
        xapi.ad_hoc(modify_qs=True,
                    qs={'type': 'version'})
        elem = xapi.element_root.find('./result/model')
        if elem is not None:
            if elem.text == 'Panorama' or elem.text.startswith('M-'):
                panorama = True
        else:
            print("Can't get model", file=sys.stderr)
    except pan.xapi.PanXapiError as e:
        print('pan.xapi.PanXapi:', e, file=sys.stderr)
        sys.exit(1)

    try:
        xpath = Xpath(panorama=panorama,
                      vsys=args.vsys,
                      template=args.template)
    except ValueError as e:
        print(e, file=sys.stderr)
        sys.exit(1)

    if args.debug > 1:
        print('Xpath():', str(xpath), file=sys.stderr)
        print(xpath.certificates(), file=sys.stderr)
        print(xpath.trusted_root_ca(), file=sys.stderr)
        print(xpath.root_ca_exclude_list(), file=sys.stderr)

    if args.show:
        show(xapi, xpath)
    if args.show_tree:
        show_tree(xapi, xpath)

    # experimental
    if args.enable_trusted:
        enable_trusted(xapi, xpath)

    # experimental
    if args.disable_trusted:
        disable_trusted(xapi, xpath)

    if args.update_trusted:
        update_trusted_root_cas(xapi, xpath, quiet=False)

    if args.delete:
        delete_certs(xapi, xpath)

    if args.update:
        if args.certs is None:
            print('--certs argument required', file=sys.stderr)
            sys.exit(1)
        if args.type is None:
            print('--type argument required', file=sys.stderr)
            sys.exit(1)

        update_certs(xapi, xpath)

    if args.commit:
        commit(xapi, panorama)


def exclude_cert(sha256):
    EXCLUDE = [
        # This root is in microsoft root store and is PAN-OS predefined
        # 0216_CCA_India_2015_SPL.
        # Expires Jan 29 11:36:43 2025 GMT.
        # 'openssl verify -check_ss_sig' fails with bad signature
        # PAN-257401
        'C34C5DF53080078FFE45B21A7F600469917204F4F0293F1D7209393E5265C04F',
        # XXX Signature Algorithm: rsassaPss
        '233525D6E906A9B99176204E3C2B4FBF5CEE03F2D126B2E64428BDF97CBC6138',
    ]

    if sha256 in EXCLUDE:
        if args.debug:
            print('Skip problem certificate %s' % sha256, file=sys.stderr)
        return True

    return False


def show_disabled_trusted(xapi, xpath):
    kwargs = {'xpath': xpath.root_ca_exclude_list()}
    api_request(xapi, xapi.get, kwargs, 'success', ['7', '19'])
    disabled = xapi.element_root.findall(
        './result/root-ca-exclude-list/member')

    if len(disabled):
        print('Warning: %d Default Trusted CAs are disabled; '
              'to enable run guard.py --enable-trusted' % len(disabled))


def enable_trusted(xapi, xpath):
    kwargs = {'xpath': xpath.root_ca_exclude_list()}
    api_request(xapi, xapi.get, kwargs, 'success', ['7', '19'])
    disabled = xapi.element_root.findall(
        './result/root-ca-exclude-list/member')

    if args.dry_run:
        kwargs2 = {'xpath': '/config/predefined/trusted-root-ca'}
        api_request(xapi, xapi.get, kwargs2, 'success', '19')
        total = xapi.element_root.findall(
            './result/trusted-root-ca/entry')

        print('enable-trusted dry-run: %d default trusted root CAs,'
              ' %d are disabled, %d to enable' % (
                  len(total), len(disabled), len(disabled)))
        return

    if len(disabled):
        api_request(xapi, xapi.delete, kwargs, 'success', '20')

    print('%d default trusted root CAs enabled' % len(disabled))


def disable_trusted(xapi, xpath):
    kwargs = {'xpath': '/config/predefined/trusted-root-ca'}
    api_request(xapi, xapi.get, kwargs, 'success', '19')
    entries = xapi.element_root.findall('./result/trusted-root-ca/entry')

    kwargs = {'xpath': xpath.root_ca_exclude_list()}
    api_request(xapi, xapi.get, kwargs, 'success', ['7', '19'])
    disabled = xapi.element_root.findall(
        './result/root-ca-exclude-list/member')

    if args.dry_run:
        enabled = len(entries) - len(disabled)
        print('disable-trusted dry-run: %d default trusted root CAs,'
              ' %d are enabled, %d to disable' % (
                  len(entries), enabled, enabled))
        return

    disabled_ = []
    for entry in disabled:
        name = entry.text
        disabled_.append(name)

    members = []
    for entry in entries:
        name = entry.attrib['name']
        if name in disabled_:
            continue
        member = '<member>%s</member>' % name
        members.append(member)

        if args.verbose:
            print('disabling', name, file=sys.stderr)

    if members:
        root_ca_exclude = xpath.root_ca_exclude_list()
        element = ''.join(members)

        kwargs = {
            'xpath': root_ca_exclude,
            'element': element,
        }
        api_request(xapi, xapi.set, kwargs, 'success', '20')

    print('%d default trusted root CAs disabled' % len(members))


def get_certs(xapi, xpath):
    certificates = xpath.certificates()
    kwargs = {'xpath': certificates}
    api_request(xapi, xapi.get, kwargs, 'success', ['7', '19'])
    if xapi.status_code == '7':
        return []

    entries = xapi.element_root.findall('./result/certificate/entry')

    data = {}
    prog = re.compile(NAME_RE)
    progcn = re.compile(r'/CN=([^/]+)')

    for entry in entries:
        name = entry.attrib['name']
        if prog.search(name):
            subject = entry.find('./subject').text
            subject_cn = None
            match = progcn.search(subject)
            if match:
                subject_cn = match.group(1)
            issuer = entry.find('./issuer').text
            issuer_cn = None
            match = progcn.search(issuer)
            if match:
                issuer_cn = match.group(1)
            expiry = entry.find('./expiry-epoch').text
            try:
                if time.time() > int(expiry):
                    expired = True
                else:
                    expired = False
            except ValueError as e:
                print('%s expiry-epoch %s: %s' % (
                    name, expiry, e), file=sys.stderr)

            v = {
                'cert-name': name,
                'subject': subject,
                'subject-cn': subject_cn,
                'subject-hash': entry.find('./subject-hash').text,
                'issuer': issuer,
                'issuer-cn': issuer_cn,
                'issuer-hash': entry.find('./issuer-hash').text,
                'expired': expired,
            }
            data[name] = v

    return data


def delete_certs(xapi, xpath):
    data = get_certs(xapi, xpath)

    if args.dry_run:
        print('delete dry-run: %d to delete' % len(data))
        return

    for name in data:
        delete_cert(xapi, xpath, name)

    print('%d certificates deleted' % len(data))


def delete_cert(xapi, xpath, name):
    rootca = xpath.trusted_root_ca()
    member = "/member[text()='%s']" % name
    kwargs = {'xpath': rootca + member}
    api_request(xapi, xapi.delete, kwargs, 'success', ['7', '20'])

    certificates = xpath.certificates()
    entry = "/entry[@name='%s']" % name
    kwargs = {'xpath': certificates + entry}
    # XXX can return status_code 7 intermittently; workaround is
    # to re-run
    api_request(xapi, xapi.delete, kwargs, 'success', '20')
    if args.verbose:
        print('deleted', name, file=sys.stderr)


def get_trusted_root_cas(xapi, xpath):
    trusted_root_ca = xpath.trusted_root_ca()
    kwargs = {'xpath': trusted_root_ca}
    api_request(xapi, xapi.get, kwargs, 'success', ['7', '19'])

    prog = re.compile(NAME_RE)
    data = []
    if xapi.status_code == '19':
        entries = xapi.element_root.findall(
            './result/trusted-root-CA/member')
        for entry in entries:
            name = entry.text
            if prog.search(name):
                data.append(name)

    return data


def update_trusted_root_cas(xapi, xpath, quiet=True):
    cert_names = get_certs(xapi, xpath)
    add = []

    if cert_names:
        data = get_trusted_root_cas(xapi, xpath)
        if data:
            for name in cert_names:
                if name not in data:
                    add.append(name)
        else:
            add.extend(cert_names)

    if args.dry_run:
        print('update-trusted dry-run: %d certificates'
              ' to enable as trusted root CAs' % len(add))
        return

    if add:
        add_trusted_root_cas(xapi, xpath, add)

    if not quiet:
        print('%d certificates enabled as trusted root CA' % len(add))


def add_trusted_root_cas(xapi, xpath, cert_names):
    members = []

    for x in cert_names:
        members.append('<member>%s</member>' % x)

    trusted_root_ca = xpath.trusted_root_ca()
    element = ''.join(members)

    kwargs = {
        'xpath': trusted_root_ca,
        'element': element,
    }

    api_request(xapi, xapi.set, kwargs, 'success', '20')


def update_certs(xapi, xpath):
    old = get_certs(xapi, xpath)

    new = {}
    try:
        data = pan_chainguard.util.read_cert_archive(
            path=args.certs)
    except pan_chainguard.util.UtilError as e:
        print(str(e), file=sys.stderr)
        sys.exit(1)

    for sha256 in data:
        if exclude_cert(sha256):
            continue
        cert_type, content = data[sha256]
        if cert_type not in args.type:
            continue
        cert_name = pan_chainguard.util.hash_to_name(sha256=sha256)
        new[cert_name] = content

    old_set = set(old)
    new_set = set(new.keys())

    if args.debug > 2:
        print('old', pprint.pformat(old_set), file=sys.stderr)
        print('new', pprint.pformat(new_set), file=sys.stderr)

    delete = list(old_set - new_set)
    add = list(new_set - old_set)

    if args.debug > 1:
        print('update delete', pprint.pformat(delete), file=sys.stderr)
        print('update add', pprint.pformat(add), file=sys.stderr)

    if args.dry_run:
        print('update dry-run: %d to delete, %d to add' % (
            len(delete), len(add)))
        return

    for name in delete:
        delete_cert(xapi, xpath, name)
    print('%d certificates deleted' % len(delete))

    total = 0
    for name in add:
        if add_cert(xapi, xpath, name, new[name]):
            total += 1

    if total:
        add_trusted_root_cas(xapi, xpath, add_cert.cert_names)

    print('%d certificates added' % total)


def add_cert(xapi, xpath, cert_name, content):
    kwargs = {
        'category': 'certificate',
        'file': content,
        'extra_qs': {
            'certificate-name': cert_name,
            'format': 'pem',
        },
    }

    if xpath.panorama:
        if args.template is not None:
            kwargs['extra_qs']['target-tpl'] = args.template
        if args.vsys is not None:
            # XXX does not work; PAN-257229
            kwargs['extra_qs']['target-tpl-vsys'] = args.vsys
    elif args.vsys is not None:
        kwargs['vsys'] = args.vsys

    try:
        xapi.import_file(**kwargs)
    except pan.xapi.PanXapiError as e:
        if 'Certificate is expired' in str(e):
            if args.verbose:
                print('certificate is expired %s' % cert_name, file=sys.stderr)
            return False
        else:
            print('%s: %s: %s' % (xapi.import_file.__name__, kwargs, e),
                  file=sys.stderr)
            sys.exit(1)

    # Use function attribute to cache certificate names so we can use
    # a single API request to enable them as trusted root CAs.
    try:
        add_cert.cert_names.append(cert_name)
    except AttributeError:
        add_cert.cert_names = [cert_name]

    if args.verbose:
        print('added %s' % cert_name, file=sys.stderr)

    return True


def show(xapi, xpath):
    data = get_certs(xapi, xpath)
    out = []
    num_expired = 0

    for x in data:
        expired = ''
        if data[x]['expired']:
            num_expired += 1
            expired = ' (expired)'
        if data[x]['subject'] == data[x]['issuer']:
            # Root
            # XXX subject == issuer can be an intermediate
            type_ = 'R'
        else:
            # Intermediate
            type_ = 'I'
        m = '%s %s%s Subject: "%s" Issuer: "%s"' % (
            x, type_, expired,
            data[x]['subject'], data[x]['issuer'])
        out.append(m)

    expired = ''
    if num_expired:
        expired = ' (%d expired)' % num_expired
    print('%d Device Certificates%s' % (len(out), expired))
    if out and args.verbose:
        print('\n'.join(out))

    if out:
        data = get_trusted_root_cas(xapi, xpath)
        num = len(data)
        print('%d Trusted Root CA Certificates' % num)
        if num < len(out):
            print('Warning: %d certificates not trusted; '
                  'run guard.py --update-trusted' % (len(out) - num))

    show_disabled_trusted(xapi, xpath)


def duplicates_in_path(tree: Tree) -> list[dict]:
    def dfs(nid: str, seen: set[str], out: list[dict]) -> None:
        new_seen = set(seen)
        node = tree.get_node(nid)
        if node.tag != 'Root':
            key = node.data['subject-hash']

            if key in seen:
                out.append({
                    'node-id': nid,
                    'tag': str(node.tag),
                    'key': key,
                    'node-data': node.data,
                })
                return  # stop this root-to-leaf path here

            else:
                new_seen.add(key)

        children = tree.children(nid)
        if not children:
            return  # leaf reached, no dup on this path

        for child in children:
            dfs(child.identifier, new_seen, out)

    duplicates = []
    dfs(tree.root, set(), duplicates)

    return duplicates


def show_tree(xapi, xpath):
    data = get_certs(xapi, xpath)
    issuers = defaultdict(list)
    subjects = defaultdict(list)
    roots = []

    for x in data:
        issuers[data[x]['issuer-hash']].append(data[x])
        subjects[data[x]['subject-hash']].append(data[x])
        if data[x]['subject-hash'] == data[x]['issuer-hash']:
            roots.append(data[x])

    orphans = []

    for x in data:
        if data[x]['issuer-hash'] not in subjects:
            orphans.append(data[x])

    if args.debug:
        print('issuers', pprint.pformat(issuers), file=sys.stderr)
        print('subjects', pprint.pformat(subjects), file=sys.stderr)
        print('roots', pprint.pformat(roots), file=sys.stderr)
        print('orphans', pprint.pformat(orphans), file=sys.stderr)

    tree = Tree()
    tree.add_node(Node(tag='Root', identifier=0))

    def build_node(x):
        subject = x['subject-cn'] if x['subject-cn'] else x['subject']
        issuer = x['issuer-cn'] if x['issuer-cn'] else x['issuer']
        tag = (f"{x['cert-name']} "
               f'Subject: "{subject}" '
               f'Issuer: "{issuer}"')

        node = Node(tag=tag, identifier=x['cert-name'], data=x)

        return node

    def add_children(parent, issuer):
        if issuer in issuers:
            for x in issuers[issuer]:
                if (tree.contains(x['cert-name']) or
                   x['subject-hash'] == x['issuer-hash']):
                    continue

                tree.add_node(build_node(x), parent=parent)
                add_children(x['cert-name'], x['subject-hash'])

    def format_stats(tree):
        stats = pan_chainguard.util.stats_from_tree(tree=tree)

        for k, v in stats.items():
            name = k.replace('_', ' ').title()
            value = '%.4f' % v if isinstance(v, float) else '%d' % v
            print('%s: %s' % (name, value))

    for x in roots + orphans:
        tree.add_node(build_node(x), parent=0)
        add_children(x['cert-name'], x['subject-hash'])

    print(tree.show(stdout=False), end='')

    if args.verbose:
        format_stats(tree)

    duplicates = duplicates_in_path(tree)
    if args.verbose and duplicates:
        print(f'Info: {len(duplicates)} duplicate subject in tree path')
        for x in duplicates:
            print(f"{x['key']} {x['tag']}")

    treelen = len(tree) - 1  # don't count root node
    sublen = 0
    for x in subjects:
        sublen += len(subjects[x])
    if treelen != sublen:
        print('Error: tree length != number of certificates:'
              ' %d != %d' % (treelen, sublen))

    duplicates = []
    for x in subjects:
        if len(subjects[x]) > 1:
            duplicates.append(subjects[x][0]['subject'])

    if args.verbose and duplicates:
        print('Info: %d duplicate certificate subjects' %
              len(duplicates))
        print('\n'.join(duplicates))


def commit(xapi, panorama):
    root = etree.Element('commit')
    partial = etree.SubElement(root, 'partial')
    desc = etree.SubElement(partial, 'description')
    desc.text = '%s %s' % (title, __version__)
    if args.admin:
        admin = etree.SubElement(partial, 'admin')
        admin_member = etree.SubElement(admin, 'member')
        admin_member.text = args.admin

    policy_and_objects = etree.Element('policy-and-objects')
    policy_and_objects.text = 'excluded'
    device_and_network = etree.Element('device-and-network')
    device_and_network.text = 'excluded'
    shared_object = etree.Element('shared-object')
    shared_object.text = 'excluded'

    if panorama:
        if args.template:
            # commit scope: template
            # no template vsys scope
            template = etree.SubElement(partial, 'template')
            template_member = etree.SubElement(template, 'member')
            template_member.text = args.template
        else:
            # commit scope: device-and-network
            partial.append(shared_object)
    else:
        # firewall
        if args.vsys:
            # commit scope: vsys
            vsys = etree.SubElement(partial, 'vsys')
            vsys_member = etree.SubElement(vsys, 'member')
            vsys_member.text = args.vsys
            partial.append(device_and_network)
            partial.append(shared_object)
        else:
            # commit scope: shared-object
            partial.append(device_and_network)
            partial.append(policy_and_objects)

    cmd = etree.tostring(root).decode()
    if args.debug:
        print(cmd, file=sys.stderr)

    kwargs = {
        'cmd': cmd,
        'sync': True,
    }

    if args.dry_run:
        return

    if args.verbose:
        print('commit config for admin %s' % args.admin)

    api_request(xapi, xapi.commit, kwargs, 'success')
    if args.debug:
        print(xapi.xml_root(), file=sys.stderr)

    if xapi.status_code is not None:
        code = ' [code=\"%s\"]' % xapi.status_code
    else:
        code = ''
    if xapi.status is not None:
        print('commit: %s%s' % (xapi.status, code), end='')
    if args.verbose and xapi.status_detail is not None:
        print(': "%s"' % xapi.status_detail.rstrip(), end='')
    print()


def api_request(xapi, func, kwargs, status=None, status_code=None):
    try:
        func(**kwargs)
    except pan.xapi.PanXapiError as e:
        print('%s: %s: %s' % (func.__name__, kwargs, e),
              file=sys.stderr)
        sys.exit(1)

    status_detail = (' "%s"' % xapi.status_detail
                     if xapi.status_detail is not None
                     else '')
    if (status is not None and
       not pan_chainguard.util.s1_in_s2(xapi.status, status)):
        print('%s: %s:%s status %s != %s' %
              (func.__name__, kwargs, xapi.status_detail,
               xapi.status, status),
              file=sys.stderr)
        sys.exit(1)

    if (status_code is not None and
       not pan_chainguard.util.s1_in_s2(xapi.status_code, status_code)):
        print('%s: %s:%s status_code %s != %s' %
              (func.__name__, kwargs, status_detail,
               xapi.status_code, status_code),
              file=sys.stderr)
        sys.exit(1)


def parse_args():
    def check_vsys(x):
        if x.isdigit():
            x = 'vsys' + x
        return x

    parser = argparse.ArgumentParser(
        usage='%(prog)s [options]',
        description='update PAN-OS trusted CAs')
    parser.add_argument('--tag', '-t',
                        required=True,
                        help='.panrc tagname')
    parser.add_argument('--vsys',
                        type=check_vsys,
                        help='vsys name or number')
    parser.add_argument('--template',
                        help='Panorama template')
    parser.add_argument('--certs',
                        metavar='PATH',
                        help='certificate archive path')
    parser.add_argument('--update',
                        action='store_true',
                        help='update certificates')
    parser.add_argument('--delete',
                        action='store_true',
                        help='delete all previously added certificates')
    parser.add_argument('-T', '--type',
                        action='append',
                        choices=['root', 'intermediate'],
                        help='certificate type(s) for update')
    # XXX experimental
    # We don't want to have a condition where default trusted CAs
    # are disabled but there are no replacement root CAs.  Better
    # to just keep default trusted CAs enabled.
    parser.add_argument('--disable-trusted',
                        action='store_true',
                        help=argparse.SUPPRESS)
#                        help='disable all default trusted root CAs')
    parser.add_argument('--enable-trusted',
                        action='store_true',
                        help=argparse.SUPPRESS)
#                        help='enable all default trusted root CAs')
    parser.add_argument('--update-trusted',
                        action='store_true',
                        help='update trusted root CA for all certificates')
    parser.add_argument('--commit',
                        action='store_true',
                        help='commit configuration')
    parser.add_argument('--dry-run',
                        action='store_true',
                        help="don't update PAN-OS")
    parser.add_argument('--show',
                        action='store_true',
                        help='show %s managed config' % title)
    parser.add_argument('--show-tree',
                        action='store_true',
                        help='show %s managed certificates in tree format' %
                        title)
    parser.add_argument('--admin',
                        help='commit admin')
    parser.add_argument('--xdebug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='pan.xapi debug')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='enable verbosity')
    parser.add_argument('--debug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='enable debug')
    x = '%s %s' % (title, __version__)
    parser.add_argument('--version',
                        action='version',
                        help='display version',
                        version=x)
    args = parser.parse_args()

    if args.debug:
        print(args, file=sys.stderr)

    return args


if __name__ == '__main__':
    main()



#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import csv
import sys
from typing import Optional


class MozillaError(Exception):
    pass


# https://wiki.mozilla.org/CA/Intermediate_Certificates
#
# Intermediate CA Certificates:
# https://ccadb.my.salesforce-sites.com/mozilla/PublicAllIntermediateCertsWithPEMCSV
#

# Non-revoked, non-expired Intermediate CA Certificates chaining up
# to roots in Mozilla's program with the Websites trust bit set:
# https://ccadb.my.salesforce-sites.com/mozilla/MozillaIntermediateCertsCSVReport

class MozillaCaCerts:
    def __init__(self, *,
                 path: str):
        def insert(row):
            self.certs[row[SHA256]] = row[PEM]

        self.certs = {}
        self.name = None

        try:
            with open(path, 'r', newline='') as csvfile:
                reader = csv.DictReader(csvfile,
                                        dialect='unix')
                row = next(reader)
                if 'SHA256' in row and 'PEM' in row:
                    SHA256 = 'SHA256'
                    PEM = 'PEM'
                    self.name = 'MozillaIntermediateCerts'
                elif ('SHA-256 Fingerprint' in row and
                      'PEM Info' in row):
                    SHA256 = 'SHA-256 Fingerprint'
                    PEM = 'PEM Info'
                    self.name = 'PublicAllIntermediateCerts'
                else:
                    raise MozillaError('Invalid CSV')
                insert(row)

                for row in reader:
                    insert(row)

        except OSError as e:
            raise MozillaError(str(e))

    def get_cert_pem(self, *, sha256: str) -> Optional[str]:
        if sha256 not in self.certs:
            return

        pem = self.certs[sha256]

        # XXX PublicAllIntermediateCertsWithPEMReport.csv has
        # single quotes around the PEM.
        c = "'"
        if pem.startswith(c):
            pem = pem[1:]
        if pem.endswith(c):
            pem = pem[:-1]

        return pem


# https://wiki.mozilla.org/CA/Intermediate_Certificates
#
# The following reports list the intermediate certificates that have
# been added to OneCRL, and their revocation status as indicated by
# the CA in the CCADB:
# https://ccadb.my.salesforce-sites.com/mozilla/IntermediateCertsInOneCRLReportCSV

class MozillaOneCrl:
    def __init__(self, *,
                 path: str,
                 debug: bool = False):
        self.certs = {}

        try:
            with open(path, 'r', newline='') as csvfile:
                reader = csv.DictReader(csvfile,
                                        dialect='unix')
                for row in reader:
                    sha256 = row['SHA-256 Fingerprint']
                    if sha256 in self.certs and debug:
                        print('Duplicate in OneCRL %s' % (
                            sha256), file=sys.stderr)
                    self.certs[sha256] = row

        except OSError as e:
            raise MozillaError(str(e))

    def get(self, *, sha256: str) -> Optional[dict[str, str]]:
        if sha256 not in self.certs:
            return

        return self.certs[sha256]



#!/usr/bin/env python3

#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import argparse
import asyncio
from collections import defaultdict
import csv
import json
import os
import sys
from treelib import Node, Tree

libpath = os.path.dirname(os.path.abspath(__file__))
sys.path[:0] = [os.path.join(libpath, os.pardir)]

from pan_chainguard import title, __version__
from pan_chainguard.ccadb import *
from pan_chainguard.mozilla import MozillaOneCrl, MozillaError
import pan_chainguard.util

args = None


def main():
    global args
    args = parse_args()

    ret = asyncio.run(main_loop())

    sys.exit(ret)


async def main_loop():
    onecrl = None
    if args.onecrl:
        try:
            onecrl = MozillaOneCrl(path=args.onecrl)
        except MozillaError as e:
            print('MozillaOneCrl: %s' % e, file=sys.stderr)
            sys.exit(1)

    certs, invalid, warning = get_certs(onecrl)

    tree = get_tree(certs, invalid)
    if args.debug > 2:
        x = tree.show(stdout=False)
        print(x, file=sys.stderr, end='')

    total_invalid, newtree, intermediates = get_intermediates(
        tree, invalid, warning)

    if args.debug:
        x = newtree.show(stdout=False)
        print(x, file=sys.stderr, end='')

    write_fingerprints(intermediates)
    write_tree(newtree)

    return 0


def get_certs(onecrl):
    invalid = {}
    warning = {}
    certs = {}
    duplicates = defaultdict(list)

    try:
        with open(args.ccadb, 'r', newline='') as csvfile:
            reader = csv.DictReader(csvfile,
                                    dialect='unix')
            for row in reader:
                sha256 = row['SHA-256 Fingerprint']
                name = row['Certificate Name']
                cert_type = row['Certificate Record Type']
                parent_sha256 = row['Parent SHA-256 Fingerprint']

                ret, err = revoked(row)
                if ret:
                    x = '%s %s %s' % (err, sha256, name)
                    if args.debug > 2:
                        print(x, file=sys.stderr)
                    invalid[sha256] = x
                    continue

                ret, err = valid_from_to(row)
                if not ret:
                    x = '%s %s %s' % (err, sha256, name)
                    if args.debug > 2:
                        print(x, file=sys.stderr)
                    invalid[sha256] = x
                    continue

                # For duplicate certificate fingerprints in CCADB,
                # retain a root certificate, or intermediate when no
                # root.
                if sha256 in certs:
                    if sha256 not in duplicates:
                        duplicates[sha256].append(certs[sha256])
                    duplicates[sha256].append(row)

                    if (certs[sha256]['Certificate Record Type'] ==
                       'Root Certificate'):
                        if args.debug > 1:
                            print('Retain duplicate %s %s' % (
                                  certs[sha256]['Certificate Record Type'],
                                  sha256), file=sys.stderr)
                        continue
                    elif cert_type == 'Root Certificate':
                        if args.debug > 1:
                            print('Replace duplicate %s'
                                  ' with %s %s' % (
                                      certs[sha256]['Certificate Record Type'],
                                      cert_type, sha256), file=sys.stderr)
                    else:
                        if args.debug > 1:
                            print('Skip duplicate %s, retain %s %s' % (
                                cert_type,
                                certs[sha256]['Certificate Record Type'],
                                sha256), file=sys.stderr)
                        continue

                if cert_type == 'Root Certificate':
                    if parent_sha256:
                        x = 'Root with parent: %s' % sha256
                        if args.debug > 1:
                            print(x, file=sys.stderr)
                        invalid[sha256] = x
                        continue

                if cert_type == 'Intermediate Certificate':
                    if onecrl:
                        r = onecrl.get(sha256=sha256)
                        if r is not None:
                            x = 'In OneCRL %s "%s" "%s"' % (
                                sha256, name, r['Revocation Status'])
                            if r['Comments']:
                                x += ' "%s"' % (
                                    r['Comments'].replace('\r\n', ' '))
                            if args.debug > 1:
                                print(x, file=sys.stderr)
                            invalid[sha256] = x
                            continue

                    if not parent_sha256:
                        x = 'Intermediate with no parent: %s' % sha256
                        if args.debug > 1:
                            print(x, file=sys.stderr)
                        invalid[sha256] = x
                        continue

                    trust_bits = derived_trust_bits(row)
                    if TrustBits.SERVER_AUTHENTICATION not in trust_bits:
                        x = 'Missing %s in %s %s' % (
                            TrustBits.SERVER_AUTHENTICATION.name,
                            trust_bits, sha256)
                        if args.debug > 1:
                            print(x, file=sys.stderr)
                        invalid[sha256] = x
                        continue

                certs[sha256] = row

    except OSError as e:
        print('%s: %s' % (args.ccadb, e), file=sys.stderr)
        sys.exit(1)

    if args.debug > 1:
        for sha256 in duplicates:
            print('Duplicate certificates %s' % sha256, file=sys.stderr)
            for x in duplicates[sha256]:
                print('  %s %s %s' % (x['Certificate Name'],
                                      x['Certificate Record Type'],
                                      x['Salesforce Record ID']),
                      file=sys.stderr)

    return certs, invalid, warning


def get_tree(certs, invalid):
    nodes = {}

    for row in certs.values():
        sha256 = row['SHA-256 Fingerprint']
        name = row['Certificate Name']
        cert_type = row['Certificate Record Type']
        parent_name = row['Parent Certificate Name']

        tag = f'{sha256} Subject: "{name}"'
        if cert_type == 'Root Certificate':
            tag += f' CA-Owner: "{parent_name}"'
        else:
            tag += f' Issuer: "{parent_name}"'
        nodes[sha256] = Node(tag=tag, identifier=sha256, data=row)

    waiting_nodes = defaultdict(list)

    tree = Tree()
    tree.add_node(Node(tag='Root', identifier=0))

    for node in nodes.values():
        row = node.data
        sha256 = row['SHA-256 Fingerprint']
        parent_sha256 = row['Parent SHA-256 Fingerprint']

        if not parent_sha256:
            tree.add_node(nodes[sha256], parent=0)
        else:
            if tree.contains(parent_sha256):
                tree.add_node(nodes[sha256], parent=nodes[parent_sha256])
                # is node a parent for any waiting nodes
                if sha256 in waiting_nodes:
                    for x in waiting_nodes[sha256]:
                        tree.add_node(nodes[x], parent=nodes[sha256])
                    del waiting_nodes[sha256]
            elif parent_sha256 not in invalid:
                # node's parent not in tree, add node to waiting list
                waiting_nodes[parent_sha256].append(sha256)

    if args.debug > 1 and waiting_nodes:
        print('Warning: nodes with no parent', file=sys.stderr)
        for x in waiting_nodes:
            print('parent', x, 'nodes', waiting_nodes[x], file=sys.stderr)

    return tree


def get_intermediates(tree, invalid, warning):
    intermediates = []
    total_invalid = 0

    EXCLUDE_INTERMEDIATES = [
        'C0A6F4DC63A24BFDCF54EF2A6A082A0A72DE35803E2FF5FF527AE5D87206DFD5',
    ]

    try:
        data = pan_chainguard.util.read_fingerprints(
            path=args.root_fingerprints)
    except pan_chainguard.util.UtilError as e:
        print('%s: %s' % (args.root_fingerprints, e), file=sys.stderr)
        sys.exit(1)

    newtree = Tree()
    newtree.add_node(Node(tag='Root', identifier=0))

    for row in data:
        sha256 = row['sha256']

        if sha256 in warning:
            print('Certificate warning: %s' % (
                warning[sha256]), file=sys.stderr)

        if sha256 in invalid:
            print('Invalid certificate: %s' % (
                invalid[sha256]), file=sys.stderr)
            total_invalid += 1
            continue

        if not tree.contains(sha256):
            print('Not found in CCADB: %s' % (
                sha256), file=sys.stderr)
            total_invalid += 1
            continue

        node = tree.get_node(sha256)
        status_root = node.data['Status of Root Cert']
        statuses = status_root.split(';')
        included = [': Included' in x for x in statuses]
        if not any(included):
            print('Certificate not in common root store: '
                  '%s: %s' % (sha256, status_root),
                  file=sys.stderr)

        if sha256 in EXCLUDE_INTERMEDIATES:
            print('Skip intermediates for root: %s' % (
                sha256), file=sys.stderr)
            newtree.add_node(node=node, parent=0)
            continue

        subtree = tree.subtree(sha256)
        nodes = subtree.all_nodes()
        for node in nodes[1:]:
            intermediates.append(node.identifier)
        for node in nodes:
            parent = subtree.parent(node.identifier)
            newtree.add_node(
                node=node,
                parent=0 if parent is None else parent.identifier)

        if args.debug > 1:
            print(subtree, end='', file=sys.stderr)
            nodes_ = [x.identifier for x in nodes]
            print(len(nodes_), nodes_, file=sys.stderr)

    return total_invalid, newtree, intermediates


def write_fingerprints(intermediates):
    if not args.int_fingerprints:
        return

    data = []
    for x in intermediates:
        row = {
            'type': 'intermediate',
            'sha256': x,
        }
        data.append(row)

    try:
        pan_chainguard.util.write_fingerprints(
            path=args.int_fingerprints,
            data=data)
    except pan_chainguard.util.UtilError as e:
        print('%s: %s' % (args.int_fingerprints, e), file=sys.stderr)
        sys.exit(1)

    if args.verbose:
        print('%d total intermediate certificates' % len(data))


def write_tree(tree):
    if not args.tree:
        return

    data = pan_chainguard.util.tree_to_dict(tree=tree)

    try:
        with open(args.tree, 'w') as f:
            json.dump(data, f, separators=(',', ':'))
    except (OSError, TypeError) as e:
        print('%s: %s' % (args.tree, e), file=sys.stderr)
        sys.exit(1)


def parse_args():
    parser = argparse.ArgumentParser(
        usage='%(prog)s [options]',
        description='determine intermediate CAs')
    # https://ccadb.my.salesforce-sites.com/ccadb/AllCertificateRecordsCSVFormatv3
    parser.add_argument('-c', '--ccadb',
                        required=True,
                        metavar='PATH',
                        help='CCADB all certificate information CSV path')
    parser.add_argument('-r', '--root-fingerprints',
                        required=True,
                        metavar='PATH',
                        help='root CA fingerprints CSV path')
    # https://ccadb.my.salesforce-sites.com/mozilla/IntermediateCertsInOneCRLReportCSV
    parser.add_argument('-o', '--onecrl',
                        metavar='PATH',
                        help='Mozilla OneCRL CSV path')
    parser.add_argument('-i', '--int-fingerprints',
                        metavar='PATH',
                        help='intermediate CA fingerprints CSV path')
    parser.add_argument('--tree',
                        metavar='PATH',
                        help='save certificate tree as JSON to path')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='enable verbosity')
    parser.add_argument('--debug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='enable debug')
    x = '%s %s' % (title, __version__)
    parser.add_argument('--version',
                        action='version',
                        help='display version',
                        version=x)
    args = parser.parse_args()

    if args.debug:
        print(args, file=sys.stderr)

    return args


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

#
# Copyright (c) 2024 Palo Alto Networks, Inc.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import aiohttp
import argparse
import asyncio
import os
import sys
import time

libpath = os.path.dirname(os.path.abspath(__file__))
sys.path[:0] = [os.path.join(libpath, os.pardir)]

from pan_chainguard import title, __version__
from pan_chainguard.ccadb import *
from pan_chainguard.crtsh import ArgsError, CrtShApi
import pan_chainguard.mozilla
import pan_chainguard.util

MAX_TASKS = 3  # concurrent crt.sh API requests
CRT_SH_TIMEOUT = 60

args = None


def main():
    global args
    args = parse_args()

    ret = asyncio.run(main_loop())

    sys.exit(ret)


async def main_loop():
    if not pan_chainguard.util.is_writable(args.certs_new):
        print('%s: not writable' % args.certs_new, file=sys.stderr)
        sys.exit(1)

    fingerprints = []
    for x in args.fingerprints:
        try:
            fingerprints.extend(
                pan_chainguard.util.read_fingerprints(path=x))
        except pan_chainguard.util.UtilError as e:
            print(str(e), file=sys.stderr)
            sys.exit(1)

    mozilla = []
    if args.certs_mozilla:
        for x in args.certs_mozilla:
            try:
                mozilla.append(
                    pan_chainguard.mozilla.MozillaCaCerts(path=x))
            except pan_chainguard.mozilla.MozillaError as e:
                print(str(e), file=sys.stderr)
                sys.exit(1)

    certs_old = {}
    if args.certs_old:
        try:
            certs_old = pan_chainguard.util.read_cert_archive(
                path=args.certs_old)
        except pan_chainguard.util.UtilError as e:
            print(str(e), file=sys.stderr)
            sys.exit(1)

    errors, certs = await get_certs(fingerprints, mozilla, certs_old)

    try:
        pan_chainguard.util.write_cert_archive(
            path=args.certs_new, data=certs)
    except pan_chainguard.util.UtilError as e:
        print(str(e), file=sys.stderr)
        sys.exit(1)

    if args.verbose:
        print('Total certs-new: %d' % len(certs))

    return 2 if errors else 0


async def get_certs(fingerprints, mozilla, certs_old):
    certs = {}
    crtsh_download = {}
    total_certs_old = 0
    total_mozilla = {
        'MozillaIntermediateCerts': 0,
        'PublicAllIntermediateCerts': 0,
    }
    total_crtsh = 0

    for row in fingerprints:
        cert_type = row['type']
        sha256 = row['sha256']

        # may be duplicates in multiple fingerprint files
        if sha256 in certs:
            continue

        if sha256 in certs_old:
            total_certs_old += 1
            certs[sha256] = certs_old[sha256]
            continue

        if cert_type == 'intermediate':
            for mozilla_certs in mozilla:
                pem = mozilla_certs.get_cert_pem(sha256=sha256)
                if pem is not None:
                    total_mozilla[mozilla_certs.name] += 1
                    certs[sha256] = (cert_type, pem)
                    break
            if sha256 in certs:
                continue

        crtsh_download[sha256] = cert_type

    errors, certificates = await get_cert_files(crtsh_download.keys())
    for sha256 in certificates:
        total_crtsh += 1
        certs[sha256] = (crtsh_download[sha256], certificates[sha256])

    if args.verbose:
        print('certs-old: %d' % total_certs_old)
        for x in total_mozilla:
            print('%s: %d' % (x, total_mozilla[x]))
        print('crt.sh: %d' % total_crtsh)

    return errors, certs


async def download(api, sha256):
    tries = 0
    MAX_TRIES = 5
    RETRY_SLEEP = 5.0

    RETRY_STATUS = [
        429,  # Too Many Requests
        502,  # Bad Gateway
        503,  # Service Unavailable
        504,  # Gateway Time-out
    ]

    while True:
        tries += 1
        try:
            if args.verbose and tries == 1:
                print('Download using crt.sh API %s' % sha256)

            resp = await api.download(id=sha256)

            if resp.status in RETRY_STATUS:
                x = '%d %s %s' % (
                    resp.status, resp.reason, sha256)
                if tries == MAX_TRIES:
                    print('no retry after try %d %s' % (tries, x),
                          file=sys.stderr)
                    x = 'download failed ' + x
                    break
                print('retry after try %d %s' % (tries, x),
                      file=sys.stderr)
                await asyncio.sleep(RETRY_SLEEP)
            elif resp.status != 200:
                x = 'download failed %d %s %s' % (
                    resp.status, resp.reason, sha256)
                break
            else:
                filename, content = await api.content(resp=resp)

                start = '-----BEGIN CERTIFICATE-----'
                end = '-----END CERTIFICATE-----\n'
                if (content.startswith(start) and
                   content.endswith(end)):
                    return content, sha256
                else:
                    # XXX ephemeral?
                    x = 'content malformed %s %s' % (
                        sha256, content)
                    if tries == MAX_TRIES:
                        print('no retry after try %d %s' % (tries, x),
                              file=sys.stderr)
                        x = 'download failed ' + x
                        break
                    print('retry after try %d %s' % (tries, x),
                          file=sys.stderr)
                    await asyncio.sleep(RETRY_SLEEP)

        except (asyncio.TimeoutError,
                asyncio.CancelledError,  # XXX
                aiohttp.ClientError) as e:
            msg = e if str(e) else type(e).__name__
            x = 'CrtShApi: %s %s' % (msg, sha256)
            if tries == MAX_TRIES:
                print('no retry after try %d %s' % (tries, x),
                      file=sys.stderr)
                x = 'download failed ' + x
                break
            print('retry after try %d %s' % (tries, x),
                  file=sys.stderr)
            # no sleep

        except ArgsError as e:
            x = 'CrtShApi: %s' % e
            break

    return None, x


async def get_cert_files(sha256):
    user_agent = '%s/%s' % (title, __version__)
    headers = {'user-agent': user_agent}

    try:
        api = CrtShApi(timeout=CRT_SH_TIMEOUT, headers=headers)
    except ArgsError as e:
        print('CrtShApi: %s' % e, file=sys.stderr)
        sys.exit(1)
    else:
        errors, certificates = await download_certs(api, sha256)
    finally:
        await api.session.close()

    return errors, certificates


async def download_certs(api, sha256):
    certificates = {}
    tasks = []
    start = time.time()
    total = 0
    errors = 0

    for x in sha256:
        total += 1
        tasks.append(download(api, x))

        if len(tasks) == MAX_TASKS:
            errors += await run_tasks(tasks, certificates)
            tasks = []

    if len(tasks):
        errors += await run_tasks(tasks, certificates)

    if args.debug:
        end = time.time()
        elapsed = end - start
        mins = elapsed // 60
        secs = elapsed % 60
        avg = elapsed / total if total > 0 else 0
        print('%d tasks, elapsed %.2f seconds, %dm%ds, avg %.2fs' %
              (total, elapsed, mins, secs, avg),
              file=sys.stderr)

    return errors, certificates


async def run_tasks(tasks, certificates):
    INTERVAL_WAIT = 3.3  # random throttle sweet spot to avoid TimeoutError
    errors = 0

    if args.debug:
        print('running %d tasks' % len(tasks),
              file=sys.stderr)
        start = time.time()

    for coro in asyncio.as_completed(tasks):
        content, sha256 = await coro
        if content is None:
            # error
            print(sha256, file=sys.stderr)
            errors += 1
        else:
            certificates[sha256] = content

    if args.debug:
        end = time.time()
        elapsed = end - start
        rate = len(tasks) / elapsed
        rate2 = elapsed / len(tasks)
        print('%d tasks complete, elapsed %.2f seconds, '
              '%.2f tasks/sec, %.2f secs/task' %
              (len(tasks), elapsed, rate, rate2), file=sys.stderr)
        print('sleeping %.2fs' % INTERVAL_WAIT, file=sys.stderr)
        await asyncio.sleep(INTERVAL_WAIT)

    return errors


def parse_args():
    parser = argparse.ArgumentParser(
        usage='%(prog)s [options]',
        description='get CA certificates')
    parser.add_argument('-f', '--fingerprints',
                        required=True,
                        action='append',
                        metavar='PATH',
                        help='CA fingerprints CSV path')
    parser.add_argument('-m', '--certs-mozilla',
                        action='append',
                        metavar='PATH',
                        help='Mozilla certs with PEM CSV path')
    parser.add_argument('--certs-old',
                        metavar='PATH',
                        help='old certificate archive path')
    parser.add_argument('--certs-new',
                        required=True,
                        metavar='PATH',
                        help='new certificate archive path')
    parser.add_argument('--verbose',
                        action='store_true',
                        help='enable verbosity')
    parser.add_argument('--debug',
                        type=int,
                        choices=[0, 1, 2, 3],
                        default=0,
                        help='enable debug')
    x = '%s %s' % (title, __version__)
    parser.add_argument('--version',
                        action='version',
                        help='display version',
                        version=x)
    args = parser.parse_args()

    if args.debug:
        print(args, file=sys.stderr)

    return args


if __name__ == '__main__':
    main()
